from os import environ
from pathlib import Path
from re import match

from email_validator import validate_email
from requests import get
from sqlalchemy import ForeignKey, UniqueConstraint, inspect
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates
from ..validators import validate_directory
from ..utils import get_format_string_vars
from ..base import Base
from ..custom_types import (
    SamplesheetString,
    StrippedString,
    int_pk,
    stripped_str,
    unique_stripped_str,
)


class Institution(Base, kw_only=True):
    __tablename__ = 'institution'

    id: Mapped[int_pk] = mapped_column(init=False, repr=False, compare=False)
    email_format: Mapped[stripped_str] = mapped_column(repr=False, compare=False)
    name: Mapped[unique_stripped_str] = mapped_column(default=None, index=True)
    short_name: Mapped[stripped_str] = mapped_column(
        default=None, index=True, compare=False
    )
    country: Mapped[str] = mapped_column(StrippedString(length=2), default='US')
    state: Mapped[str | None] = mapped_column(StrippedString(length=2), default=None)
    city: Mapped[stripped_str] = mapped_column(default=None)
    ror_id: Mapped[unique_stripped_str | None] = mapped_column(
        default=None, repr=False, compare=False
    )

    @validates('email_format')
    def check_email_format(self, key: str, email_format: str) -> str:
        email_format = email_format.strip().lower()

        variables = get_format_string_vars(email_format)

        if not variables:
            raise ValueError(f'No variables found in email format {email_format}')

        person_columns = set(inspect(Person).columns.keys())
        non_existent_person_columns = variables - person_columns

        if non_existent_person_columns:
            raise ValueError(
                f'The following variables in {email_format} are not members of {person_columns}:\n{non_existent_person_columns}'
            )

        example_values = {var: 'string' for var in variables}
        example_email = email_format.format_map(example_values)

        validate_email(example_email)

        return email_format

    @validates('ror_id')
    def check_ror_id(self, key: str, ror_id: str | None) -> str | None:
        if ror_id is None:
            return ror_id

        ror_id = ror_id.strip()
        base_url = 'https://api.ror.org/organizations'
        url = f'{base_url}/{ror_id}'
        response = get(url)

        if not response.ok:
            raise ValueError(
                f'ROR ID {ror_id} not found in database search of {base_url}'
            )

        data = response.json()

        if self.name is None:
            self.name = data['name']

        acronyms = data['acronyms']
        aliases = data['aliases']
        if self.short_name is None:
            if len(acronyms) > 0:
                self.short_name = acronyms[0]
            elif len(aliases) > 0:
                self.short_name = aliases[0]
            else:
                pass

        self.country = data['country']['country_code']

        addresses = data['addresses']
        if len(addresses) > 0:
            geonames_city_info = addresses[0]['geonames_city']
            self.city = geonames_city_info['city']
            if self.country == 'US':
                state_code = geonames_city_info['geonames_admin1']['code']
                _, self.state = state_code.split('.')
        else:
            raise ValueError(
                f'Could not find city information from ROR for {self.name}. Please enter manually.'
            )

        return ror_id


class Person(Base, kw_only=True):
    __tablename__ = 'person'

    id: Mapped[int_pk] = mapped_column(init=False, repr=False, compare=False)
    first_name: Mapped[stripped_str]
    last_name: Mapped[stripped_str]
    # name: Mapped[stripped_str] = mapped_column(init=False, default=None, index=True)

    institution_id: Mapped[int] = mapped_column(
        ForeignKey('institution.id'), repr=False, init=False, compare=False
    )
    institution: Mapped[Institution] = relationship(repr=False, compare=False)

    email_auto_generated: Mapped[bool] = mapped_column(
        init=False, default=False, repr=False, compare=False
    )
    email: Mapped[unique_stripped_str] = mapped_column(default=None, index=True)
    orcid: Mapped[unique_stripped_str | None] = mapped_column(
        default=None, repr=False, compare=False
    )

    @validates('first_name', 'last_name')
    def format_name(self, key: str, name: str) -> str:
        formatted_split = name.strip().title().split()
        noramlized_inner_whitespace = ' '.join(formatted_split)
        return noramlized_inner_whitespace

    @validates('orcid')
    def check_orcid(self, key: str, orcid: str | None) -> str | None:
        if orcid is None:
            return orcid

        orcid = orcid.strip()

        orcid_pattern = r'^(\d{4})-?(\d{4})-?(\d{4})-?(\d{4}|\d{3}X)$'
        if (match_obj := match(orcid_pattern, string=orcid)) is None:
            raise ValueError(f'ORCID {orcid} does not match pattern {orcid_pattern}')

        digit_groups = match_obj.groups()
        formatted_orcid = '-'.join(digit_groups)

        base_url = 'https://pub.orcid.org'
        url = f'{base_url}/{formatted_orcid}'
        headers = {'Accept': 'application/json'}
        response = get(url, headers=headers)

        if not response.ok:
            raise ValueError(
                f'{formatted_orcid} not found in database search of {base_url}'
            )

        return formatted_orcid

    @validates('email')
    def check_email(self, key: str, email: str | None) -> str | None:
        if email is None:
            variables = get_format_string_vars(self.institution.email_format)
            var_values = {var: getattr(self, var) for var in variables}

            email = self.institution.email_format.format_map(var_values).replace(' ', '')
        
        email_info = validate_email(email.lower(), check_deliverability=True)
        return email_info.normalized


class Lab(Base, kw_only=True):
    __tablename__ = 'lab'

    id: Mapped[int_pk] = mapped_column(init=False, repr=False, compare=False)

    institution_id: Mapped[int] = mapped_column(
        ForeignKey('institution.id'), init=False, repr=False, compare=False
    )
    pi_id: Mapped[int] = mapped_column(
        ForeignKey('person.id'), init=False, repr=False, compare=False
    )

    institution: Mapped[Institution] = relationship()
    pi: Mapped[Person] = relationship()

    name: Mapped[stripped_str] = mapped_column(default=None, index=True)
    delivery_dir: Mapped[unique_stripped_str] = mapped_column(default=None, repr=False)
    unix_group: Mapped[stripped_str] = mapped_column(
        init=False, default=None, repr=False, compare=False
    )

    __table_args__ = (UniqueConstraint('institution_id', 'pi_id', 'name'),)

    @validates('name')
    def set_name(self, key: str, name: str | None) -> str:
        # This assumes that no two PIs share the same name and
        # institution. Might have to change in production
        return (
            f'{self.pi.first_name} {self.pi.last_name} Lab'
            if name is None
            else name.title()
        )

    @validates('delivery_dir')
    def set_delivery_dir(self, key: str, delivery_dir: str | None) -> str:
        # Getting delivery parent dir from environment for the sake of
        # testing. Hopefully can figure out a better way, importing from
        # defaults instead of this
        delivery_parent_dir = Path(environ['delivery_parent_dir'])

        # If no delivery directory is provided, get it automatically
        # from PI name
        if delivery_dir is None:
            pi = self.pi

            first_name = SamplesheetString().process_bind_param(
                pi.first_name.lower(), dialect=None
            )
            last_name = SamplesheetString().process_bind_param(
                pi.last_name.lower(), dialect=None
            )

            delivery_path = Path(f'{first_name}_{last_name}')
        else:
            delivery_path = Path(delivery_dir)

        validate_directory(delivery_parent_dir, required_structure={delivery_path: []})

        return str(delivery_parent_dir / delivery_path)

    @validates('unix_group')
    def set_group(self, key: str, unix_group: None) -> str:  # type: ignore
        return Path(self.delivery_dir).group()



from re import sub
from string import ascii_letters, digits
from typing import Annotated

from sqlalchemy.orm import mapped_column
from sqlalchemy.types import Integer, String, TypeDecorator


class StrippedString(TypeDecorator):
    """
    A string type that strips whitespace from the ends of the string
    before sending to the database.
    """

    impl = String
    cache_ok = True

    def process_bind_param(self, string: str | None, dialect) -> str | None:
        return None if string is None else string.strip()


class SamplesheetString(TypeDecorator):
    """
    A string type that removes illegal characters from the string before
    sending to the database. Suitable for strings that will be used in
    the samplesheet passed as input to the `nf-tenx` pipeline.
    """

    impl = String
    cache_ok = True

    def process_bind_param(self, string: str | None, dialect) -> str | None:
        if string is None:
            return None

        sep_chars = r'[\s_-]'
        samplesheet_blacklist_pattern = rf'[^{ascii_letters + digits + sep_chars}]'

        # Remove illegal characters
        string = sub(
            pattern=samplesheet_blacklist_pattern, repl='', string=string.strip()
        )

        # Replace separator characters with hyphens
        string = sub(pattern=sep_chars, repl='-', string=string)

        # Replace occurrences of multiple hyphens with a single hyphen
        string = sub(pattern=r'-+', repl='-', string=string)

        return string


class XeniumSlideSerialNumber(TypeDecorator):
    """ """

    impl = Integer
    cache_ok = True

    def process_bind_param(self, serial_number: str, dialect) -> int:
        return int(serial_number.strip())

    def process_result_value(self, serial_number: int, dialect) -> str:
        return f'{serial_number:02}'


# Commonly used primary key types
int_pk = Annotated[int, mapped_column(primary_key=True)]
stripped_str_pk = Annotated[str, mapped_column(StrippedString, primary_key=True)]
samplesheet_str_pk = Annotated[str, mapped_column(SamplesheetString, primary_key=True)]

# Commonly used non-unique string types
stripped_str = Annotated[str, mapped_column(StrippedString)]
samplesheet_str = Annotated[str, mapped_column(SamplesheetString)]

# Commonly used unique types
unique_int = Annotated[int, mapped_column(unique=True)]
unique_stripped_str = Annotated[str, mapped_column(StrippedString, unique=True)]
unique_samplesheet_str = Annotated[str, mapped_column(SamplesheetString, unique=True)]
xenium_slide_serial_number = Annotated[
    str, mapped_column(XeniumSlideSerialNumber, unique=True)
]



from dataclasses import field
from datetime import date
from re import match

from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates

from ..base import Base
from ..custom_types import (
    SamplesheetString,
    samplesheet_str,
    samplesheet_str_pk,
    stripped_str,
    stripped_str_pk,
)
from .entities import Lab, Person


class Platform(Base, kw_only=True):
    __tablename__ = 'platform'

    # Platform attributes
    name: Mapped[stripped_str_pk]
    data_set_id_prefix: Mapped[str] = mapped_column(SamplesheetString(length=2))
    sample_id_prefix: Mapped[str] = mapped_column(SamplesheetString(length=2))
    data_set_id_length: Mapped[int]
    sample_id_length: Mapped[int]

    @validates('data_set_id_prefix', 'sample_id_prefix')
    def check_prefix(self, key: str, prefix: str) -> str:
        pattern = r'^[A-Z]{2}$'
        prefix = prefix.upper().strip()

        if match(pattern, prefix) is None:
            raise ValueError(
                f'[green]{key}[/] [orange1]{prefix}[/] does not match the pattern [green]{pattern}[/].'
            )

        return prefix

    @validates('data_set_id_length', 'sample_id_length')
    def check_id_length(self, key: str, id_length: int) -> int:
        min_id_length = 7
        max_id_length = 9

        if not min_id_length <= id_length <= max_id_length:
            raise ValueError(
                f'[green]{key}[/] must be between {min_id_length} and {max_id_length}, but {id_length} was given.'
            )

        return id_length

    @property
    def prefixes(self) -> dict[str, str]:
        return {'DataSet': self.data_set_id_prefix, 'Sample': self.sample_id_prefix}

    @property
    def id_lengths(self) -> dict[str, int]:
        return {'DataSet': self.data_set_id_length, 'Sample': self.sample_id_length}


class DataSet(Base, kw_only=True):
    __tablename__ = 'data_set'

    # DataSet attributes
    # TODO: auto-incrementing behavior
    id: Mapped[samplesheet_str_pk]
    name: Mapped[samplesheet_str] = mapped_column(index=True)
    ilab_request_id: Mapped[stripped_str] = mapped_column(
        index=True
    )  # TODO: ilab request ID validation
    date_initialized: Mapped[date] = mapped_column(repr=False)

    # Parent foreign keys
    lab_id: Mapped[int] = mapped_column(ForeignKey('lab.id'), init=False, repr=False)
    platform_name: Mapped[str] = mapped_column(ForeignKey('platform.name'), init=False)
    submitter_id: Mapped[int] = mapped_column(
        ForeignKey('person.id'), init=False, repr=False
    )

    # Parent models
    lab: Mapped[Lab] = relationship()
    platform: Mapped[Platform] = relationship()
    submitter: Mapped[Person] = relationship()

    # Automatically set attributes
    batch_id: Mapped[int] = mapped_column(init=False, default=None, repr=False)

    # Model metadata
    id_based_on: str = field(default='date_received', init=False, repr=False)

    __mapper_args__ = {
        'polymorphic_on': 'platform_name',
    }

    # TODO: implement validation
    @validates('id')
    def check_id(self, key: str, id: str) -> str:
        return id.upper().strip()

    @validates('batch_id')
    def set_batch_id(self, key: str, batch_id: None) -> int:
        # If it's decided that more things constitute a batch, this will
        # be easy to update.

        # Note that submitter email, institution, name, and ORCID have
        # been picked instead of the person ID because a person is not
        # assigned an ID until they enter the database.

        # There is a small likelihood that two people with the same name
        # and institution both have an autogenerated email and no ORCID.
        # If these two people submit on the same day, the batch_id of
        # the two datasets will be the same.
        # TODO: this should be fixed somehow
        to_hash = (
            self.date_initialized,
            self.submitter.email,
            self.submitter.institution_id,
            self.submitter.first_name,
            self.submitter.last_name,
            self.submitter.orcid,
        )
        return hash(to_hash)


class Sample(Base, kw_only=True):
    __tablename__ = 'sample'

    # Sample attributes
    id: Mapped[samplesheet_str_pk]
    name: Mapped[samplesheet_str] = mapped_column(index=True)
    date_received: Mapped[date] = mapped_column(default_factory=date.today)

    # Parent foreign keys
    data_set_id: Mapped[int] = mapped_column(
        ForeignKey('data_set.id'), init=False, repr=False
    )
    platform_name: Mapped[str] = mapped_column(ForeignKey('platform.name'), init=False)

    # Model metadata
    id_based_on: str = field(default='date_received', init=False, repr=False)

    __mapper_args__ = {'polymorphic_on': 'platform_name'}

    # TODO: implement validation
    @validates('id')
    def check_id(self, key: str, id: str) -> str:
        return id.upper().strip()



from collections.abc import Iterable, Mapping
from pathlib import Path
type DirectoryStructure = Mapping[
    str | Path, Iterable[str | Path] | 'DirectoryStructure'
] | Iterable[str | Path]
def validate_directory(
    directory: str | Path, required_structure: DirectoryStructure
) -> None:
    directory = Path(directory)

    if not directory.is_dir():
        raise NotADirectoryError(f'{directory} is not a directory.')

    if isinstance(required_structure, Mapping):
        for sub_dirname, sub_structure in required_structure.items():
            validate_directory(
                directory / sub_dirname, required_structure=sub_structure
            )
        return

    if isinstance(required_structure, Iterable):
        missing_files = {
            str(directory / filename)
            for filename in required_structure
            if not (directory / filename).is_file()
        }
        if missing_files:
            raise FileNotFoundError(
                f'The following files are missing from {directory}: {missing_files}'
            )
        return

    raise TypeError(
        f'{required_structure} must be a [green]Mapping[/] or an [green]Iterable[/], not a [orange1]{type(required_structure).__name__}[/]'
    )
